package dev.fyloz.colorrecipesexplorer.databasemanager

import liquibase.Contexts
import liquibase.Liquibase
import liquibase.database.DatabaseFactory
import liquibase.database.jvm.JdbcConnection
import liquibase.resource.ClassLoaderResourceAccessor
import mu.KotlinLogging
import org.slf4j.Logger
import java.sql.Connection
import java.sql.DriverManager

internal const val DATABASE_VERSION_NOT_INITIALIZED = 0

internal const val METADATA_TABLE_NAME = "updater_metadata"
internal const val COLUMN_KEY_NAME = "metadata_key"
internal const val COLUMN_VALUE_NAME = "metadata_value"

internal const val ROW_VERSION_NAME = "version"

internal const val QUERY_SELECT_VERSION =
    "SELECT $COLUMN_VALUE_NAME FROM $METADATA_TABLE_NAME WHERE $COLUMN_KEY_NAME='$ROW_VERSION_NAME'"


/** [CreDatabase]'s context. Contains data needed by the utility like the [properties] and the [logger]. */
data class DatabaseContext(
    val properties: DatabaseUpdaterProperties,
    val logger: Logger
)

/** DSL for [DatabaseContext]. */
fun databaseContext(
    properties: DatabaseUpdaterProperties,
    logger: Logger = KotlinLogging.logger { },
    op: DatabaseContext.() -> Unit = {}
) = DatabaseContext(properties, logger).apply(op)


class CreDatabase(
    context: DatabaseContext
) {
    private val properties = context.properties
    private val logger = context.logger

    /** Connection to the database server */
    val serverConnection = tryConnect(properties.url)

    private var databaseConnectionOpen = false

    /** Connection to the database */
    val databaseConnection by lazy {
        tryConnect(properties.url + properties.dbName)
            .apply { databaseConnectionOpen = true }
    }

    private var databaseOpen = false
    val database by lazy {
        DatabaseFactory.getInstance().findCorrectDatabaseImplementation(JdbcConnection(databaseConnection))
            .apply { databaseOpen = true }
    }

    /** If the database exists */
    val exists: Boolean by lazy {
        with(serverConnection.metaData.catalogs) {
            while (next()) {
                if (properties.dbName == getString(1)) {
                    return@with true
                }
            }
            // If no catalog has the name of the targeted database, it does not exist
            false
        }
    }

    /** If the updater's metadata table exists in the database */
    val metadataTableExists by lazy {
        with(
            serverConnection.metaData.getTables(properties.dbName, null, METADATA_TABLE_NAME, arrayOf("TABLE"))
        ) {
            // If there is no result, the table does not exist
            this.next()
        }
    }

    /** The version of the database */
    val version by lazy { fetchVersion() }

    /** Updates the database to the target version. */
    fun update() {
        val targetVersion = properties.targetVersion

        if (version >= targetVersion) {
            logger.warn("Database version is superior or equals to the target version. No changes will be made.")
            return
        }

        logger.info("Updating database from version $version to $targetVersion...")

        val liquibase =
            Liquibase("/changelogs/changelog.master.$targetVersion.xml", ClassLoaderResourceAccessor(), database)
        liquibase.update(Contexts())
        logger.debug("Update to version $targetVersion completed!")
        liquibase.close()

        logger.info("Update completed!")
    }

    /** Gets the database version. */
    private fun fetchVersion(): Int {
        fun getVersionFromDatabase(): Int {
            val statement = databaseConnection.createStatement()
            val results = statement.executeQuery(QUERY_SELECT_VERSION)

            if (!results.next()) {
                logger.debug("The version row in the metadata table was not found, assuming version $DATABASE_VERSION_NOT_INITIALIZED")
                return DATABASE_VERSION_NOT_INITIALIZED
            }

            return results.getInt(COLUMN_VALUE_NAME)
        }

        // Check if the database exists
        if (!exists) {
            throw CreDatabaseException.UnknownDatabase(properties.dbName)
        }
        // Check if the metadata table exists
        if (!metadataTableExists) {
            logger.debug("The metadata table was not found, assuming version $DATABASE_VERSION_NOT_INITIALIZED")
            return DATABASE_VERSION_NOT_INITIALIZED
        }

        return getVersionFromDatabase()
    }

    /** Try to create a [Connection] with the given [url]. */
    private fun tryConnect(url: String): Connection =
        try {
            DriverManager.getConnection(url, properties.username, properties.password)
        } catch (ex: Exception) {
            throw CreDatabaseException.Connection(url, ex)
        }

    /** Closes all connections to the database. */
    fun close() {
        serverConnection.close()
        if (databaseConnectionOpen) databaseConnection.close()
    }
}

sealed class CreDatabaseException(message: String, exception: Exception? = null) : Exception(message, exception) {
    /** Thrown when an error occurs while creating a connection to the database at the given [url]. */
    class Connection(val url: String, exception: Exception) :
        CreDatabaseException("Could not create a connection to the database at '$url'", exception)

    /** Thrown when the database with the given [name] cannot be found. */
    class UnknownDatabase(val name: String) :
        CreDatabaseException("Unknown database '$name'")
}
